# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module provides hooks to be used with pre-commit.

This module provides hooks for usage with pre-commit (https://pre-commit.com).
Note that the logging is disabled for this module as pre-commit works best
in conjunction with print statements.
"""

import mistune
import re
import yaml
from . import foam
from os import path
from datetime import date
from jinja2 import Environment, FileSystemLoader


def preview_encoding(file: str) -> bool:
    """Process a file in a similar way as Rodare does for preview.

    This function uses the same code as Rodare to generated a html text for
    preview of a given file. In case the encoding with 'ascii' is not possible
    an error is returned.

    Args:
        file: file to be encoded.

    Returns:
        True, if an issue is found
    """
    try:
        with open(file, encoding="ascii") as fp:
            mistune.markdown(fp.read())
    except UnicodeDecodeError as e:
        text = e.object.decode("utf-8")
        before = text[e.start - 10 : e.start]
        after = text[e.end : e.end + 10]
        issue = text[e.start : e.end]
        print(f"{file}: contains non-ASCII characters:")
        print(f"    ... {before}>>>{issue}<<<{after} ...")
        return True

    return False


def line_length(file: str) -> bool:
    """Check line length for less than 80 characters.

    This function mimics a hook from the software released by the OpenFOAM
    Foundation and checks for the correct line length.

    Args:
        file: file to be checked.

    Returns:
        True, if an issue is found
    """
    pattern = re.compile("^.{81,}".encode())
    with open(file, "rb") as f:
        for line_no, line in enumerate(f, start=1):
            if pattern.search(line):
                print(f"{file}: {line_no}: Line longer than 80 characters.")
                return True
    return False


def header_ifndef(file: str) -> bool:
    """Check for ifndef directive or define statement.

    This function mimics a hook from the software released by the OpenFOAM
    Foundation and checks for ifndef directive and define statements.

    Correct is: #ifndef filename_H
    Not correct is: #ifndef banana_H or #if !defined filename_H

    Args:
        file: file to be checked.

    Returns:
        True, if an issue is found
    """
    fname = path.basename(file).replace(".", "_")

    # Pattern to detect #ifndef directives
    p1 = re.compile("#(ifndef|if !defined)".encode())
    # Pattern to detect correct #ifndef directive with filename
    # Note: use \n instead of ^, as the file content is read in as one string
    p2 = re.compile(rf"\n#ifndef.*[ _]{fname}".encode())
    # Pattern to detect correct define statement with filename
    # Note: use \n instead of ^, as the file content is read in as one string
    p3 = re.compile(rf"\n#define.*[ _]{fname}".encode())

    return_value = False

    with open(file, "rb") as f:
        content = f.read()
        if p1.search(content):
            # print(content)
            if not p2.search(content):
                print(
                    f"{file}: #ifndef directive should be corrected: "
                    + f"#ifndef {fname}"
                )
                return_value = True
            if not p3.search(content):
                print(
                    f"{file}: #define statement should be corrected: "
                    + f"#define {fname}"
                )
                return_value = True
    return return_value


def tabs(file: str) -> bool:
    """Check for tabs.

    This function mimics a hook from the software released by the OpenFOAM
    Foundation and checks for tabs.

    Args:
        file: file to be checked.

    Returns:
        True, if an issue is found
    """
    pattern = re.compile("\t".encode())
    with open(file, "rb") as f:
        for line_no, line in enumerate(f, start=1):
            if pattern.search(line):
                print(
                    f"{file}: {line_no}: Replace tabs with "
                    + f"{foam.indention} whitespaces."
                )
                return True
    return False


def nonstandardcode(file: str) -> bool:
    """Check for non-standard code.

    This function mimics a hook from the software released by the OpenFOAM
    Foundation and checks for non-standard code.

    Args:
        file: file to be checked.

    Returns:
        True, if an issue is found
    """
    p1 = re.compile("> >".encode())
    p2 = re.compile("NULL".encode())

    return_value = False

    with open(file, "rb") as f:
        for line_no, line in enumerate(f, start=1):
            if p1.search(line):
                print(
                    f"{file}: {line_no}: Spaced ending of multi-level "
                    + "template parameters should be: 'List<List<scalar>>'."
                )
                return_value = True
            if p2.search(line):
                print(
                    f"{file}: {line_no}: Replace 'NULL' macro with 'nullptr'."
                )
                return_value = True
    return return_value


def copyrightyear(file: str) -> bool:
    """Check for year in copyright.

    This function mimics a hook from the software released by the OpenFOAM
    Foundation and checks for copyright year.

    Args:
        file: file to be checked.

    Returns:
        True, if an issue is found
    """
    year = date.today().strftime("%Y")
    p1 = re.compile(r"Copyright\s\(C\)".encode())
    p2 = re.compile(rf"Copyright\s\(C\)\s{year}".encode())

    return_value = False

    with open(file, "rb") as f:
        for line_no, line in enumerate(f, start=1):
            if p1.search(line):
                if not p2.search(line):
                    print(f"{file}: {line_no}: Update the copyright date.")
                    return_value = True
    return return_value


def copyrightheader(file: str, pattern: str) -> bool:
    """Check for copyright header.

    This function mimics a hook from the software released by the OpenFOAM
    Foundation and checks for copyright header.

    Args:
        file: file to be checked.
        pattern: regular expression matching a valid header

    Returns:
        True, if an issue is found
    """
    pattern = re.compile(pattern.encode(), re.MULTILINE)

    with open(file, "rb") as f:
        content = f.read()
        if not pattern.search(content):
            print(f"{file}: Missing copyright header.")
            return True

    return False


def keywords(file: str, reffile: str) -> bool:
    """Check keywords against a reference list.

    This function checks the keywords in the metadata of a case against a
    given reference list. The hook takes the given file, usually a case.yml or
    a case.yml.orig, renders the content through Jinja2 and checks the keywords.

    Args:
        file: case metadata file to be checked.
        reffile: YAML file with reference keyword list

    Returns:
        True, if an issue is found
    """
    # Read file and render with Jinja2
    fpath = path.dirname(file)
    fname = path.basename(file)
    cname = path.basename(fpath)

    env = Environment(loader=FileSystemLoader(fpath))
    template = env.get_template(path.basename(fname))
    metadata = yaml.safe_load(template.render({"caseNumber": cname}))

    # Read keywords for case
    try:
        kwords = metadata[cname]["keywords"]
    except KeyError:
        print(f"{file}: No keywords specified.")
        return True

    try:
        with open(reffile) as f:
            reference = yaml.safe_load(f)
            try:
                for category in kwords.keys():
                    if kwords[category] is not None:
                        current = set(kwords[category])
                        other = set(reference[category])
                        diff = current.difference(other)
                        if diff:
                            print(
                                f"{file}: "
                                + f"Unknown keyword(s): {', '.join(diff)}."
                            )
                            return True
                    else:
                        print(
                            f"{file}: "
                            + f"Category {category} evaluates to empty."
                        )
                        return True
            except KeyError:
                print(f"{file}: Unknown category: {category}.")
                return True
    except FileNotFoundError:
        print(f"{file}: Unable to find reference list: {reffile}.")
        return True

    return False
